"""Test zha alarm control panel."""

import logging
from unittest.mock import AsyncMock, call, patch, sentinel

import pytest
from zigpy.device import Device as ZigpyDevice
from zigpy.zcl.clusters import security
import zigpy.zcl.foundation as zcl_f

from tests.common import join_zigpy_device, zigpy_device_from_json
from zha.application import Platform
from zha.application.gateway import Gateway
from zha.application.platforms.alarm_control_panel import AlarmControlPanel
from zha.application.platforms.alarm_control_panel.const import AlarmState
from zha.zigbee.device import Device

_LOGGER = logging.getLogger(__name__)


@patch(
    "zigpy.zcl.clusters.security.IasAce.client_command",
    new=AsyncMock(return_value=[sentinel.data, zcl_f.Status.SUCCESS]),
)
async def test_alarm_control_panel(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test zhaws alarm control panel platform."""
    zigpy_device: ZigpyDevice = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/frient-a-s-kepzb-110.json",
    )
    zha_device: Device = await join_zigpy_device(zha_gateway, zigpy_device)
    cluster: security.IasAce = zigpy_device.endpoints[44].out_clusters[
        security.IasAce.cluster_id
    ]
    alarm_entity: AlarmControlPanel = zha_device.platform_entities.get(
        (
            Platform.ALARM_CONTROL_PANEL,
            f"{zigpy_device.ieee}-44-{security.IasAce.cluster_id}",
        )
    )
    assert alarm_entity is not None
    assert isinstance(alarm_entity, AlarmControlPanel)

    # test that the state is STATE_ALARM_DISARMED
    assert alarm_entity.state["state"] == AlarmState.DISARMED

    # arm_away
    cluster.client_command.reset_mock()
    await alarm_entity.async_alarm_arm_away("4321")
    await zha_gateway.async_block_till_done()
    assert cluster.client_command.call_count == 2
    assert cluster.client_command.await_count == 2
    assert cluster.client_command.call_args == call(
        4,
        security.IasAce.PanelStatus.Armed_Away,
        0,
        security.IasAce.AudibleNotification.Default_Sound,
        security.IasAce.AlarmStatus.No_Alarm,
    )
    assert alarm_entity.state["state"] == AlarmState.ARMED_AWAY

    # disarm
    await reset_alarm_panel(zha_gateway, cluster, alarm_entity)

    # trip alarm from faulty code entry. First we need to arm away
    cluster.client_command.reset_mock()
    await alarm_entity.async_alarm_arm_away("4321")
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.ARMED_AWAY
    cluster.client_command.reset_mock()

    # now simulate a faulty code entry sequence
    await alarm_entity.async_alarm_disarm("0000")
    await alarm_entity.async_alarm_disarm("0000")
    await alarm_entity.async_alarm_disarm("0000")
    await zha_gateway.async_block_till_done()

    assert alarm_entity.state["state"] == AlarmState.TRIGGERED
    assert cluster.client_command.call_count == 6
    assert cluster.client_command.await_count == 6
    assert cluster.client_command.call_args == call(
        4,
        security.IasAce.PanelStatus.In_Alarm,
        0,
        security.IasAce.AudibleNotification.Default_Sound,
        security.IasAce.AlarmStatus.Emergency,
    )

    # reset the panel
    await reset_alarm_panel(zha_gateway, cluster, alarm_entity)

    # arm_home
    await alarm_entity.async_alarm_arm_home("4321")
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.ARMED_HOME
    assert cluster.client_command.call_count == 2
    assert cluster.client_command.await_count == 2
    assert cluster.client_command.call_args == call(
        4,
        security.IasAce.PanelStatus.Armed_Stay,
        0,
        security.IasAce.AudibleNotification.Default_Sound,
        security.IasAce.AlarmStatus.No_Alarm,
    )

    # reset the panel
    await reset_alarm_panel(zha_gateway, cluster, alarm_entity)

    # arm_night
    await alarm_entity.async_alarm_arm_night("4321")
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.ARMED_NIGHT
    assert cluster.client_command.call_count == 2
    assert cluster.client_command.await_count == 2
    assert cluster.client_command.call_args == call(
        4,
        security.IasAce.PanelStatus.Armed_Night,
        0,
        security.IasAce.AudibleNotification.Default_Sound,
        security.IasAce.AlarmStatus.No_Alarm,
    )

    # reset the panel
    await reset_alarm_panel(zha_gateway, cluster, alarm_entity)

    # arm from panel
    cluster.listener_event(
        "cluster_command", 1, 0, [security.IasAce.ArmMode.Arm_All_Zones, "", 0]
    )
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.ARMED_AWAY

    # reset the panel
    await reset_alarm_panel(zha_gateway, cluster, alarm_entity)

    # arm day home only from panel
    cluster.listener_event(
        "cluster_command", 1, 0, [security.IasAce.ArmMode.Arm_Day_Home_Only, "", 0]
    )
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.ARMED_HOME

    # reset the panel
    await reset_alarm_panel(zha_gateway, cluster, alarm_entity)

    # arm night sleep only from panel
    cluster.listener_event(
        "cluster_command", 1, 0, [security.IasAce.ArmMode.Arm_Night_Sleep_Only, "", 0]
    )
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.ARMED_NIGHT

    # disarm from panel with bad code
    cluster.listener_event(
        "cluster_command", 1, 0, [security.IasAce.ArmMode.Disarm, "", 0]
    )
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.ARMED_NIGHT

    # disarm from panel with bad code for 2nd time still armed
    cluster.listener_event(
        "cluster_command", 1, 0, [security.IasAce.ArmMode.Disarm, "", 0]
    )
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.TRIGGERED

    # disarm from panel with good code
    cluster.listener_event(
        "cluster_command", 1, 0, [security.IasAce.ArmMode.Disarm, "4321", 0]
    )
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.DISARMED

    # disarm when already disarmed
    cluster.listener_event(
        "cluster_command", 1, 0, [security.IasAce.ArmMode.Disarm, "4321", 0]
    )
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.DISARMED
    assert "IAS ACE already disarmed" in caplog.text

    # panic from panel
    cluster.listener_event("cluster_command", 1, 4, [])
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.TRIGGERED

    # reset the panel
    await reset_alarm_panel(zha_gateway, cluster, alarm_entity)

    # fire from panel
    cluster.listener_event("cluster_command", 1, 3, [])
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.TRIGGERED

    # reset the panel
    await reset_alarm_panel(zha_gateway, cluster, alarm_entity)

    # emergency from panel
    cluster.listener_event("cluster_command", 1, 2, [])
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.TRIGGERED

    # reset the panel
    await reset_alarm_panel(zha_gateway, cluster, alarm_entity)
    assert alarm_entity.state["state"] == AlarmState.DISARMED

    await alarm_entity.async_alarm_trigger()
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.TRIGGERED

    # reset the panel
    await reset_alarm_panel(zha_gateway, cluster, alarm_entity)
    assert alarm_entity.state["state"] == AlarmState.DISARMED

    alarm_entity._cluster_handler.code_required_arm_actions = True
    await alarm_entity.async_alarm_arm_away()
    await zha_gateway.async_block_till_done()
    assert alarm_entity.state["state"] == AlarmState.DISARMED
    assert "Invalid code supplied to IAS ACE" in caplog.text


async def reset_alarm_panel(
    zha_gateway: Gateway,
    cluster: security.IasAce,
    entity: AlarmControlPanel,
) -> None:
    """Reset the state of the alarm panel."""
    cluster.client_command.reset_mock()
    await entity.async_alarm_disarm("4321")
    await zha_gateway.async_block_till_done()
    assert entity.state["state"] == AlarmState.DISARMED
    assert cluster.client_command.call_count == 2
    assert cluster.client_command.await_count == 2
    assert cluster.client_command.call_args == call(
        4,
        security.IasAce.PanelStatus.Panel_Disarmed,
        0,
        security.IasAce.AudibleNotification.Default_Sound,
        security.IasAce.AlarmStatus.No_Alarm,
    )
    cluster.client_command.reset_mock()



"""Test ZHA device switch."""

import asyncio
import logging
import time
from unittest import mock
from unittest.mock import call, patch

import pytest
from zigpy.exceptions import ZigbeeException
import zigpy.profiles.zha
from zigpy.quirks.registry import DeviceRegistry
from zigpy.quirks.v2 import (
    DeviceAlertLevel,
    DeviceAlertMetadata,
    ExposesFeatureMetadata,
    QuirkBuilder,
)
from zigpy.quirks.v2.homeassistant import EntityType
from zigpy.quirks.v2.homeassistant.sensor import SensorDeviceClass, SensorStateClass
import zigpy.types
from zigpy.typing import UNDEFINED
from zigpy.zcl import ClusterType
from zigpy.zcl.clusters import general
from zigpy.zcl.clusters.general import Ota, PowerConfiguration
from zigpy.zcl.foundation import Status, WriteAttributesResponse
import zigpy.zdo.types as zdo_t

from tests.common import (
    SIG_EP_INPUT,
    SIG_EP_OUTPUT,
    SIG_EP_PROFILE,
    SIG_EP_TYPE,
    create_mock_zigpy_device,
    get_entity,
    join_zigpy_device,
    zigpy_device_from_json,
)
from zha.application import Platform
from zha.application.const import (
    CLUSTER_COMMAND_SERVER,
    CLUSTER_COMMANDS_CLIENT,
    CLUSTER_COMMANDS_SERVER,
    CLUSTER_TYPE_IN,
    CLUSTER_TYPE_OUT,
    UNKNOWN,
)
from zha.application.gateway import Gateway
from zha.application.platforms import PlatformEntity
from zha.application.platforms.binary_sensor import IASZone
from zha.application.platforms.light import Light
from zha.application.platforms.sensor import LQISensor, RSSISensor
from zha.application.platforms.switch import Switch
from zha.exceptions import ZHAException
from zha.zigbee.device import (
    ClusterBinding,
    DeviceFirmwareInfoUpdatedEvent,
    ZHAEvent,
    get_device_automation_triggers,
)
from zha.zigbee.group import Group


def zigpy_device(
    zha_gateway: Gateway, with_basic_cluster_handler: bool = True, **kwargs
):
    """Return a ZigpyDevice with a switch cluster."""
    in_clusters = [general.OnOff.cluster_id]
    if with_basic_cluster_handler:
        in_clusters.append(general.Basic.cluster_id)

    endpoints = {
        3: {
            SIG_EP_INPUT: in_clusters,
            SIG_EP_OUTPUT: [],
            SIG_EP_TYPE: zigpy.profiles.zha.DeviceType.ON_OFF_SWITCH,
            SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
        }
    }
    return create_mock_zigpy_device(zha_gateway, endpoints, **kwargs)


def zigpy_device_mains(zha_gateway: Gateway, with_basic_cluster_handler: bool = True):
    """Return a ZigpyDevice with a switch cluster."""
    in_clusters = [general.OnOff.cluster_id]
    if with_basic_cluster_handler:
        in_clusters.append(general.Basic.cluster_id)

    endpoints = {
        3: {
            SIG_EP_INPUT: in_clusters,
            SIG_EP_OUTPUT: [],
            SIG_EP_TYPE: zigpy.profiles.zha.DeviceType.ON_OFF_SWITCH,
            SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
        }
    }
    return create_mock_zigpy_device(
        zha_gateway,
        endpoints,
        node_descriptor=zdo_t.NodeDescriptor(
            logical_type=zdo_t.LogicalType.EndDevice,
            complex_descriptor_available=0,
            user_descriptor_available=0,
            reserved=0,
            aps_flags=0,
            frequency_band=zdo_t.NodeDescriptor.FrequencyBand.Freq2400MHz,
            mac_capability_flags=(
                zdo_t.NodeDescriptor.MACCapabilityFlags.MainsPowered
                | zdo_t.NodeDescriptor.MACCapabilityFlags.AllocateAddress
            ),
            manufacturer_code=4447,
            maximum_buffer_size=127,
            maximum_incoming_transfer_size=100,
            server_mask=11264,
            maximum_outgoing_transfer_size=100,
            descriptor_capability_field=zdo_t.NodeDescriptor.DescriptorCapability.NONE,
        ),
    )


async def _send_time_changed(zha_gateway: Gateway, seconds: int):
    """Send a time changed event."""
    await asyncio.sleep(seconds)
    await zha_gateway.async_block_till_done(wait_background_tasks=True)


@patch(
    "zha.zigbee.cluster_handlers.general.BasicClusterHandler.async_initialize",
    new=mock.AsyncMock(),
)
async def test_check_available_success(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Check device availability success on 1st try."""
    device_with_basic_cluster_handler = zigpy_device_mains(
        zha_gateway, with_basic_cluster_handler=True
    )
    zha_device = await join_zigpy_device(zha_gateway, device_with_basic_cluster_handler)
    basic_ch = device_with_basic_cluster_handler.endpoints[3].basic

    assert not zha_device.is_coordinator
    assert not zha_device.is_active_coordinator

    basic_ch.read_attributes.reset_mock()
    device_with_basic_cluster_handler.last_seen = None
    assert zha_device.available is True
    await _send_time_changed(zha_gateway, zha_device.consider_unavailable_time + 2)
    assert zha_device.available is False
    assert basic_ch.read_attributes.await_count == 0

    device_with_basic_cluster_handler.last_seen = (
        time.time() - zha_device.consider_unavailable_time - 100
    )
    _seens = [time.time(), device_with_basic_cluster_handler.last_seen]

    def _update_last_seen(*args, **kwargs):  # pylint: disable=unused-argument
        new_last_seen = _seens.pop()
        device_with_basic_cluster_handler.last_seen = new_last_seen

    basic_ch.read_attributes.side_effect = _update_last_seen

    for entity in zha_device.platform_entities.values():
        entity.emit = mock.MagicMock(wraps=entity.emit)

    # we want to test the device availability handling alone
    zha_gateway.global_updater.stop()

    # successfully ping zigpy device, but zha_device is not yet available
    await _send_time_changed(
        zha_gateway, zha_gateway._device_availability_checker.__polling_interval + 1
    )
    assert basic_ch.read_attributes.await_count == 1
    assert basic_ch.read_attributes.await_args[0][0] == ["manufacturer"]
    assert zha_device.available is False

    for entity in zha_device.platform_entities.values():
        entity.emit.assert_not_called()
        assert not entity.available
        entity.emit.reset_mock()

    # There was traffic from the device: pings, but not yet available
    await _send_time_changed(
        zha_gateway, zha_gateway._device_availability_checker.__polling_interval + 1
    )
    assert basic_ch.read_attributes.await_count == 2
    assert basic_ch.read_attributes.await_args[0][0] == ["manufacturer"]
    assert zha_device.available is False

    for entity in zha_device.platform_entities.values():
        entity.emit.assert_not_called()
        assert not entity.available
        entity.emit.reset_mock()

    # There was traffic from the device: don't try to ping, marked as available
    await _send_time_changed(
        zha_gateway, zha_gateway._device_availability_checker.__polling_interval + 1
    )
    assert basic_ch.read_attributes.await_count == 2
    assert basic_ch.read_attributes.await_args[0][0] == ["manufacturer"]
    assert zha_device.available is True
    assert zha_device.on_network is True

    for entity in zha_device.platform_entities.values():
        entity.emit.assert_called()
        assert entity.available
        entity.emit.reset_mock()

    assert "Device is not on the network, marking unavailable" not in caplog.text
    zha_device.on_network = False

    assert zha_device.available is False
    assert zha_device.on_network is False

    assert "Device is not on the network, marking unavailable" in caplog.text

    for entity in zha_device.platform_entities.values():
        entity.emit.assert_called()
        assert not entity.available
        entity.emit.reset_mock()


@patch(
    "zha.zigbee.cluster_handlers.general.BasicClusterHandler.async_initialize",
    new=mock.AsyncMock(),
)
async def test_check_available_unsuccessful(
    zha_gateway: Gateway,
) -> None:
    """Check device availability all tries fail."""

    device_with_basic_cluster_handler = zigpy_device_mains(
        zha_gateway, with_basic_cluster_handler=True
    )
    zha_device = await join_zigpy_device(zha_gateway, device_with_basic_cluster_handler)
    basic_ch = device_with_basic_cluster_handler.endpoints[3].basic

    assert zha_device.available is True
    assert basic_ch.read_attributes.await_count == 0

    device_with_basic_cluster_handler.last_seen = (
        time.time() - zha_device.consider_unavailable_time - 2
    )

    for entity in zha_device.platform_entities.values():
        entity.emit = mock.MagicMock(wraps=entity.emit)

    # we want to test the device availability handling alone
    zha_gateway.global_updater.stop()

    # unsuccessfully ping zigpy device, but zha_device is still available
    await _send_time_changed(
        zha_gateway, zha_gateway._device_availability_checker.__polling_interval + 1
    )

    assert basic_ch.read_attributes.await_count == 1
    assert basic_ch.read_attributes.await_args[0][0] == ["manufacturer"]
    assert zha_device.available is True

    for entity in zha_device.platform_entities.values():
        entity.emit.assert_not_called()
        assert entity.available
        entity.emit.reset_mock()

    # still no traffic, but zha_device is still available
    await _send_time_changed(
        zha_gateway, zha_gateway._device_availability_checker.__polling_interval + 1
    )

    assert basic_ch.read_attributes.await_count == 2
    assert basic_ch.read_attributes.await_args[0][0] == ["manufacturer"]
    assert zha_device.available is True

    for entity in zha_device.platform_entities.values():
        entity.emit.assert_not_called()
        assert entity.available
        entity.emit.reset_mock()

    # not even trying to update, device is unavailable
    await _send_time_changed(
        zha_gateway, zha_gateway._device_availability_checker.__polling_interval + 1
    )

    assert basic_ch.read_attributes.await_count == 2
    assert basic_ch.read_attributes.await_args[0][0] == ["manufacturer"]
    assert zha_device.available is False

    for entity in zha_device.platform_entities.values():
        entity.emit.assert_called()
        assert not entity.available
        entity.emit.reset_mock()


@patch(
    "zha.zigbee.cluster_handlers.general.BasicClusterHandler.async_initialize",
    new=mock.AsyncMock(),
)
async def test_check_available_no_basic_cluster_handler(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Check device availability for a device without basic cluster."""
    device_without_basic_cluster_handler = zigpy_device(
        zha_gateway, with_basic_cluster_handler=False
    )
    caplog.set_level(logging.DEBUG, logger="homeassistant.components.zha")

    zha_device = await join_zigpy_device(
        zha_gateway, device_without_basic_cluster_handler
    )

    assert zha_device.available is True

    device_without_basic_cluster_handler.last_seen = (
        time.time() - zha_device.consider_unavailable_time - 2
    )

    assert "does not have a mandatory basic cluster" not in caplog.text
    await _send_time_changed(
        zha_gateway, zha_gateway._device_availability_checker.__polling_interval + 1
    )

    assert zha_device.available is False
    assert "does not have a mandatory basic cluster" in caplog.text


async def test_device_is_active_coordinator(
    zha_gateway: Gateway,
) -> None:
    """Test that the current coordinator is uniquely detected."""

    current_coord_dev = zigpy_device(
        zha_gateway, ieee="aa:bb:cc:dd:ee:ff:00:11", nwk=0x0000
    )
    current_coord_dev.node_desc = current_coord_dev.node_desc.replace(
        logical_type=zdo_t.LogicalType.Coordinator
    )

    old_coord_dev = zigpy_device(
        zha_gateway, ieee="aa:bb:cc:dd:ee:ff:00:12", nwk=0x0000
    )
    old_coord_dev.node_desc = old_coord_dev.node_desc.replace(
        logical_type=zdo_t.LogicalType.Coordinator
    )

    # The two coordinators have different IEEE addresses
    assert current_coord_dev.ieee != old_coord_dev.ieee

    current_coordinator = await join_zigpy_device(zha_gateway, current_coord_dev)
    stale_coordinator = await join_zigpy_device(zha_gateway, old_coord_dev)

    # Ensure the current ApplicationController's IEEE matches our coordinator's
    current_coordinator.gateway.application_controller.state.node_info.ieee = (
        current_coord_dev.ieee
    )

    assert current_coordinator.is_active_coordinator
    assert not stale_coordinator.is_active_coordinator


@pytest.mark.parametrize(
    # node_info is populated with manf and model strings
    ("manf", "model", "expected_manf", "expected_model", "expected_name"),
    [
        ("RealManf", "RealModel", "RealManf", "RealModel", "RealManf RealModel"),
        (
            "RealManf",
            None,
            "RealManf",
            "Generic Zigbee Coordinator (EZSP)",
            "RealManf Generic Zigbee Coordinator (EZSP)",
        ),
        (None, "RealModel", "", "RealModel", " RealModel"),
        (
            None,
            None,
            "",
            "Generic Zigbee Coordinator (EZSP)",
            " Generic Zigbee Coordinator (EZSP)",
        ),
        (
            "Nabu Casa",
            "Home Assistant Connect ZBT-2",
            "Nabu Casa",
            "Home Assistant Connect ZBT-2",
            "Home Assistant Connect ZBT-2",
        ),
    ],
)
async def test_coordinator_info_names(
    zha_gateway: Gateway,
    manf,
    model,
    expected_manf,
    expected_model,
    expected_name,
) -> None:
    """Test that the current coordinator device is named correctly."""

    current_coord_dev = zigpy_device(
        zha_gateway, ieee="aa:bb:cc:dd:ee:ff:00:11", nwk=0x0000
    )
    current_coord_dev.node_desc = current_coord_dev.node_desc.replace(
        logical_type=zdo_t.LogicalType.Coordinator
    )

    app = current_coord_dev.application
    app.state.node_info.ieee = current_coord_dev.ieee
    app.state.node_info.manufacturer = manf
    app.state.node_info.model = model

    current_coordinator = await join_zigpy_device(zha_gateway, current_coord_dev)
    assert current_coordinator.is_active_coordinator

    assert current_coordinator.manufacturer == expected_manf
    assert current_coordinator.model == expected_model
    assert current_coordinator.name == expected_name


async def test_async_get_clusters(
    zha_gateway: Gateway,
) -> None:
    """Test async_get_clusters method."""
    zigpy_dev = zigpy_device(zha_gateway, with_basic_cluster_handler=True)
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    assert zha_device.async_get_clusters() == {
        3: {
            CLUSTER_TYPE_IN: {
                general.Basic.cluster_id: zigpy_dev.endpoints[3].in_clusters[
                    general.Basic.cluster_id
                ],
                general.OnOff.cluster_id: zigpy_dev.endpoints[3].in_clusters[
                    general.OnOff.cluster_id
                ],
            },
            CLUSTER_TYPE_OUT: {},
        }
    }


async def test_async_get_groupable_endpoints(
    zha_gateway: Gateway,
) -> None:
    """Test async_get_groupable_endpoints method."""
    zigpy_dev = zigpy_device(zha_gateway, with_basic_cluster_handler=True)
    zigpy_dev.endpoints[3].add_input_cluster(general.Groups.cluster_id)
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    assert zha_device.async_get_groupable_endpoints() == [3]


async def test_async_get_std_clusters(
    zha_gateway: Gateway,
) -> None:
    """Test async_get_std_clusters method."""
    zigpy_dev = zigpy_device(zha_gateway, with_basic_cluster_handler=True)
    zigpy_dev.endpoints[3].profile_id = zigpy.profiles.zha.PROFILE_ID
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    assert zha_device.async_get_std_clusters() == {
        3: {
            CLUSTER_TYPE_IN: {
                general.Basic.cluster_id: zigpy_dev.endpoints[3].in_clusters[
                    general.Basic.cluster_id
                ],
                general.OnOff.cluster_id: zigpy_dev.endpoints[3].in_clusters[
                    general.OnOff.cluster_id
                ],
            },
            CLUSTER_TYPE_OUT: {},
        }
    }


async def test_async_get_cluster(
    zha_gateway: Gateway,
) -> None:
    """Test async_get_cluster method."""
    zigpy_dev = zigpy_device(zha_gateway, with_basic_cluster_handler=True)
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    assert zha_device.async_get_cluster(3, general.OnOff.cluster_id) == (
        zigpy_dev.endpoints[3].on_off
    )


async def test_async_get_cluster_attributes(
    zha_gateway: Gateway,
) -> None:
    """Test async_get_cluster_attributes method."""
    zigpy_dev = zigpy_device(zha_gateway, with_basic_cluster_handler=True)
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    assert (
        zha_device.async_get_cluster_attributes(3, general.OnOff.cluster_id)
        == zigpy_dev.endpoints[3].on_off.attributes
    )

    with pytest.raises(KeyError):
        assert (
            zha_device.async_get_cluster_attributes(3, general.BinaryValue.cluster_id)
            is None
        )


async def test_async_get_cluster_commands(
    zha_gateway: Gateway,
) -> None:
    """Test async_get_cluster_commands method."""
    zigpy_dev = zigpy_device(zha_gateway, with_basic_cluster_handler=True)
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    assert zha_device.async_get_cluster_commands(3, general.OnOff.cluster_id) == {
        CLUSTER_COMMANDS_CLIENT: zigpy_dev.endpoints[3].on_off.client_commands,
        CLUSTER_COMMANDS_SERVER: zigpy_dev.endpoints[3].on_off.server_commands,
    }


async def test_write_zigbee_attribute(
    zha_gateway: Gateway,
) -> None:
    """Test write_zigbee_attribute method."""
    zigpy_dev = zigpy_device(zha_gateway, with_basic_cluster_handler=True)
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    with pytest.raises(
        ValueError,
        match="Cluster 8 not found on endpoint 3 while writing attribute 0 with value 20",
    ):
        await zha_device.write_zigbee_attribute(
            3,
            general.LevelControl.cluster_id,
            general.LevelControl.AttributeDefs.current_level.id,
            20,
        )

    cluster = zigpy_dev.endpoints[3].on_off
    cluster.write_attributes.reset_mock()

    response: WriteAttributesResponse = await zha_device.write_zigbee_attribute(
        3,
        general.OnOff.cluster_id,
        general.OnOff.AttributeDefs.start_up_on_off.id,
        general.OnOff.StartUpOnOff.PreviousValue,
    )

    assert response is not None
    assert len(response) == 1
    status_record = response[0][0]
    assert status_record.status == Status.SUCCESS

    assert cluster.write_attributes.await_count == 1
    assert cluster.write_attributes.await_args == call(
        {
            general.OnOff.AttributeDefs.start_up_on_off.id: general.OnOff.StartUpOnOff.PreviousValue
        },
        manufacturer=UNDEFINED,
    )

    cluster.write_attributes.reset_mock()
    cluster.write_attributes.side_effect = ZigbeeException
    m1 = "Failed to set attribute: value: <StartUpOnOff.PreviousValue: 255> "
    m2 = "attribute: 16387 cluster_id: 6 endpoint_id: 3"
    with pytest.raises(
        ZHAException,
        match=f"{m1}{m2}",
    ):
        await zha_device.write_zigbee_attribute(
            3,
            general.OnOff.cluster_id,
            general.OnOff.AttributeDefs.start_up_on_off.id,
            general.OnOff.StartUpOnOff.PreviousValue,
        )

    cluster = zigpy_dev.endpoints[3].on_off
    cluster.write_attributes.reset_mock()


async def test_issue_cluster_command(
    zha_gateway: Gateway,
) -> None:
    """Test issue_cluster_command method."""
    zigpy_dev = zigpy_device(zha_gateway, with_basic_cluster_handler=True)
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    with pytest.raises(
        ValueError,
        match="Cluster 8 not found on endpoint 3 while issuing command 0 with args \\[20\\]",
    ):
        await zha_device.issue_cluster_command(
            3,
            general.LevelControl.cluster_id,
            general.LevelControl.ServerCommandDefs.move_to_level.id,
            CLUSTER_COMMAND_SERVER,
            [20],
            None,
        )

    cluster = zigpy_dev.endpoints[3].on_off

    with patch("zigpy.zcl.Cluster.request", return_value=[0x5, Status.SUCCESS]):
        await zha_device.issue_cluster_command(
            3,
            general.OnOff.cluster_id,
            general.OnOff.ServerCommandDefs.on.id,
            CLUSTER_COMMAND_SERVER,
            None,
            {},
        )

        assert cluster.request.await_count == 1


async def test_async_add_to_group_remove_from_group(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test async_add_to_group and async_remove_from_group methods."""
    zigpy_dev = zigpy_device(zha_gateway, with_basic_cluster_handler=True)
    zigpy_dev.endpoints[3].add_input_cluster(general.Groups.cluster_id)
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    group: Group = zha_device.gateway.groups[0x1001]

    assert (zha_device.ieee, 3) not in group.zigpy_group.members

    await zha_device.async_add_to_group(group.group_id)

    assert (zha_device.ieee, 3) in group.zigpy_group.members

    await zha_device.async_remove_from_group(group.group_id)

    assert (zha_device.ieee, 3) not in group.zigpy_group.members

    await zha_device.async_add_endpoint_to_group(3, group.group_id)

    assert (zha_device.ieee, 3) in group.zigpy_group.members

    await zha_device.async_remove_endpoint_from_group(3, group.group_id)

    assert (zha_device.ieee, 3) not in group.zigpy_group.members

    with patch("zigpy.device.Device.add_to_group", side_effect=ZigbeeException):
        await zha_device.async_add_to_group(group.group_id)
        assert (zha_device.ieee, 3) not in group.zigpy_group.members
        assert (
            "Failed to add device '00:0d:6f:00:0a:90:69:e7' to group: 0x1001"
            in caplog.text
        )

    with patch("zigpy.endpoint.Endpoint.add_to_group", side_effect=ZigbeeException):
        await zha_device.async_add_endpoint_to_group(3, group.group_id)
        assert (zha_device.ieee, 3) not in group.zigpy_group.members
        assert (
            "Failed to add endpoint: 3 for device: '00:0d:6f:00:0a:90:69:e7' to group: 0x1001"
            in caplog.text
        )

    # add it
    assert (zha_device.ieee, 3) not in group.zigpy_group.members
    await zha_device.async_add_to_group(group.group_id)
    assert (zha_device.ieee, 3) in group.zigpy_group.members

    with patch("zigpy.device.Device.remove_from_group", side_effect=ZigbeeException):
        await zha_device.async_remove_from_group(group.group_id)
        assert (zha_device.ieee, 3) in group.zigpy_group.members
        assert (
            "Failed to remove device '00:0d:6f:00:0a:90:69:e7' from group: 0x1001"
            in caplog.text
        )

    with patch(
        "zigpy.endpoint.Endpoint.remove_from_group", side_effect=ZigbeeException
    ):
        await zha_device.async_remove_endpoint_from_group(3, group.group_id)
        assert (zha_device.ieee, 3) in group.zigpy_group.members
        assert (
            "Failed to remove endpoint: 3 for device '00:0d:6f:00:0a:90:69:e7' from group: 0x1001"
            in caplog.text
        )


async def test_async_bind_to_group(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test async_bind_to_group method."""
    zigpy_dev = zigpy_device(zha_gateway, with_basic_cluster_handler=True)
    zigpy_dev.endpoints[3].add_input_cluster(general.Groups.cluster_id)
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    zigpy_dev_remote = zigpy_device(zha_gateway, with_basic_cluster_handler=True)
    zigpy_dev_remote._ieee = zigpy.types.EUI64.convert("00:0d:7f:00:0a:90:69:e8")
    zigpy_dev_remote.endpoints[3].add_output_cluster(general.OnOff.cluster_id)
    zha_device_remote = await join_zigpy_device(zha_gateway, zigpy_dev_remote)
    assert zha_device_remote is not None

    group: Group = zha_device.gateway.groups[0x1001]

    # add a device to the group for binding
    assert (zha_device.ieee, 3) not in group.zigpy_group.members
    await zha_device.async_add_to_group(group.group_id)
    assert (zha_device.ieee, 3) in group.zigpy_group.members

    await zha_device_remote.async_bind_to_group(
        group.group_id,
        [ClusterBinding(name="on_off", type=CLUSTER_TYPE_OUT, id=6, endpoint_id=3)],
    )
    assert (
        "0xb79c: Bind_req 00:0d:7f:00:0a:90:69:e8, ep: 3, cluster: 6 to group: 0x1001 completed: [<Status.SUCCESS: 0>]"
        in caplog.text
    )

    await zha_device_remote.async_unbind_from_group(
        group.group_id,
        [ClusterBinding(name="on_off", type=CLUSTER_TYPE_OUT, id=6, endpoint_id=3)],
    )

    m1 = "0xb79c: Unbind_req 00:0d:7f:00:0a:90:69:e8, ep: 3, cluster: 6"
    assert f"{m1} to group: 0x1001 completed: [<Status.SUCCESS: 0>]" in caplog.text


async def test_device_automation_triggers(
    zha_gateway: Gateway,
) -> None:
    """Test device automation triggers."""
    zigpy_dev = zigpy_device(zha_gateway, with_basic_cluster_handler=True)
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    assert get_device_automation_triggers(zha_device) == {
        ("device_offline", "device_offline"): {"device_event_type": "device_offline"}
    }

    assert zha_device.device_automation_commands == {}
    assert zha_device.device_automation_triggers == {
        ("device_offline", "device_offline"): {"device_event_type": "device_offline"}
    }


async def test_device_properties(
    zha_gateway: Gateway,
) -> None:
    """Test device properties."""
    zigpy_dev = zigpy_device(zha_gateway, with_basic_cluster_handler=True)
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    assert zha_device.is_mains_powered is False
    assert zha_device.is_end_device is True
    assert zha_device.is_router is False
    assert zha_device.is_coordinator is False
    assert zha_device.is_active_coordinator is False
    assert zha_device.device_type == "EndDevice"
    assert zha_device.power_source == "Battery or Unknown"
    assert zha_device.available is True
    assert zha_device.on_network is True
    assert zha_device.last_seen is not None
    assert zha_device.last_seen < time.time()

    assert zha_device.ieee == zigpy_dev.ieee
    assert zha_device.nwk == zigpy_dev.nwk
    assert zha_device.manufacturer_code == 0x1037
    assert zha_device.name == "FakeManufacturer FakeModel"
    assert zha_device.manufacturer == "FakeManufacturer"
    assert zha_device.model == "FakeModel"
    assert zha_device.is_groupable is False

    assert zha_device.power_configuration_ch is None
    assert zha_device.basic_ch is not None
    assert zha_device.firmware_version is None

    assert len(zha_device.platform_entities) == 3

    lqi_entity = zha_device.platform_entities[
        Platform.SENSOR, "00:0d:6f:00:0a:90:69:e7-3-0-lqi"
    ]
    assert type(lqi_entity) is LQISensor

    rssi_entity = zha_device.platform_entities[
        Platform.SENSOR, "00:0d:6f:00:0a:90:69:e7-3-0-rssi"
    ]
    assert type(rssi_entity) is RSSISensor

    switch_entity = zha_device.platform_entities[
        Platform.SWITCH, "00:0d:6f:00:0a:90:69:e7-3-6"
    ]
    assert isinstance(switch_entity, Switch)

    with pytest.raises(KeyError, match="Entity foo not found"):
        zha_device.get_platform_entity("bar", "foo")

    # test things are none when they aren't returned by Zigpy
    zigpy_dev.node_desc = None
    delattr(zha_device, "manufacturer_code")
    delattr(zha_device, "is_mains_powered")
    delattr(zha_device, "device_type")
    delattr(zha_device, "is_router")
    delattr(zha_device, "is_end_device")
    delattr(zha_device, "is_coordinator")

    assert zha_device.manufacturer_code is None
    assert zha_device.is_mains_powered is None
    assert zha_device.device_type is UNKNOWN
    assert zha_device.is_router is None
    assert zha_device.is_end_device is None
    assert zha_device.is_coordinator is None


async def test_device_firmware_version_syncing(zha_gateway: Gateway) -> None:
    """Test device firmware version syncing."""
    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/philips-sml001.json",
    )

    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    # Register a callback to listen for device updates
    update_callback = mock.Mock()
    zha_device.on_event(DeviceFirmwareInfoUpdatedEvent.event_type, update_callback)

    # The firmware version is restored on device initialization
    assert zha_device.firmware_version == "0x42006bb7"

    # If we update the entity, the device updates as well
    update_entity = get_entity(zha_device, platform=Platform.UPDATE)
    update_entity._ota_cluster_handler.cluster.update_attribute(
        Ota.AttributeDefs.current_file_version.id,
        zigpy.types.uint32_t(0xABCD1234),
    )

    assert zha_device.firmware_version == "0xabcd1234"

    # Duplicate updates are ignored
    update_entity._ota_cluster_handler.cluster.update_attribute(
        Ota.AttributeDefs.current_file_version.id,
        zigpy.types.uint32_t(0xABCD1234),
    )

    assert zha_device.firmware_version == "0xabcd1234"
    assert update_callback.mock_calls == [
        call(
            DeviceFirmwareInfoUpdatedEvent(
                old_firmware_version="0x42006bb7",
                new_firmware_version="0xabcd1234",
            )
        )
    ]


async def test_quirks_v2_device_renaming(zha_gateway: Gateway) -> None:
    """Test quirks v2 device renaming."""
    registry = DeviceRegistry()

    (
        QuirkBuilder("CentraLite", "3405-L", registry=registry)
        .friendly_name(manufacturer="Lowe's", model="IRIS Keypad V2")
        .add_to_registry()
    )

    zigpy_dev = registry.get_device(
        await zigpy_device_from_json(
            zha_gateway.application_controller,
            "tests/data/devices/centralite-3405-l.json",
        )
    )

    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)
    assert zha_device.model == "IRIS Keypad V2"
    assert zha_device.manufacturer == "Lowe's"


async def test_quirks_v2_device_alerts(zha_gateway: Gateway) -> None:
    """Test quirks v2 device alerts."""

    # Normal device, no alerts
    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/ikea-of-sweden-tradfri-bulb-e26-opal-1000lm.json",
    )
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)
    assert not zha_device.device_alerts

    # Explicit alerts
    registry = DeviceRegistry()

    (
        QuirkBuilder("CentraLite", "3405-L", registry=registry)
        .device_alert(level=DeviceAlertLevel.WARNING, message="Test warning")
        .add_to_registry()
    )

    zigpy_dev = registry.get_device(
        await zigpy_device_from_json(
            zha_gateway.application_controller,
            "tests/data/devices/centralite-3405-l.json",
        )
    )

    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)
    assert zha_device.device_alerts == (
        DeviceAlertMetadata(level=DeviceAlertLevel.WARNING, message="Test warning"),
    )


@pytest.mark.parametrize(
    ("json_path", "primary_platform", "primary_entity_type"),
    [
        # Light bulb
        (
            "tests/data/devices/ikea-of-sweden-tradfri-bulb-gu10-ws-400lm.json",
            Platform.LIGHT,
            Light,
        ),
        # Night light with a bulb and a motion sensor
        (
            "tests/data/devices/third-reality-inc-3rsnl02043z.json",
            Platform.LIGHT,
            Light,
        ),
        # Door sensor
        (
            "tests/data/devices/centralite-3320-l.json",
            Platform.BINARY_SENSOR,
            IASZone,
        ),
        # Smart plug with energy monitoring
        (
            "tests/data/devices/innr-sp-234.json",
            Platform.SWITCH,
            Switch,
        ),
        # Atmosphere sensor with humidity, temperature, and pressure
        (
            "tests/data/devices/lumi-lumi-weather.json",
            None,
            None,
        ),
    ],
)
async def test_primary_entity_computation(
    json_path: str,
    primary_platform: Platform | None,
    primary_entity_type: PlatformEntity | None,
    zha_gateway: Gateway,
) -> None:
    """Test primary entity computation."""

    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        json_path,
    )
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    # There is a single light entity
    primary = [e for e in zha_device.platform_entities.values() if e.primary]

    if primary_platform is None:
        assert not primary
    else:
        assert primary == [
            get_entity(zha_device, primary_platform, entity_type=primary_entity_type)
        ]


async def test_quirks_v2_primary_entity(zha_gateway: Gateway) -> None:
    """Test quirks v2 primary entity."""
    registry = DeviceRegistry()

    (
        QuirkBuilder("CentraLite", "3405-L", registry=registry)
        .sensor(
            attribute_name=PowerConfiguration.AttributeDefs.battery_quantity.id,
            cluster_id=PowerConfiguration.cluster_id,
            translation_key="battery_quantity",
            fallback_name="Battery quantity",
            primary=True,
        )
        .add_to_registry()
    )

    zigpy_dev = registry.get_device(
        await zigpy_device_from_json(
            zha_gateway.application_controller,
            "tests/data/devices/centralite-3405-l.json",
        )
    )

    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    (primary,) = [e for e in zha_device.platform_entities.values() if e.primary]
    assert primary.translation_key == "battery_quantity"


async def test_quirks_v2_prevent_default_entities(zha_gateway: Gateway) -> None:
    """Test quirks v2 can prevent creating default entities."""
    registry = DeviceRegistry()

    (
        QuirkBuilder("CentraLite", "3405-L", registry=registry)
        .prevent_default_entity_creation(endpoint_id=123)
        .prevent_default_entity_creation(cluster_id=0x4567)
        .prevent_default_entity_creation(unique_id_suffix="_something")
        .prevent_default_entity_creation(function=lambda entity: None)
        .prevent_default_entity_creation(
            function=lambda entity: entity.__class__.__name__ == "IdentifyButton"
        )
        .add_to_registry()
    )

    zigpy_dev = registry.get_device(
        await zigpy_device_from_json(
            zha_gateway.application_controller,
            "tests/data/devices/centralite-3405-l.json",
        )
    )

    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    with pytest.raises(KeyError):
        zha_device.get_platform_entity(
            Platform.BUTTON, unique_id="00:0d:6f:00:05:65:83:f2-1-3"
        )

    assert len(zha_device.platform_entities) == 7


async def test_quirks_v2_change_entity_metadata(zha_gateway: Gateway) -> None:
    """Test quirks v2 can change entity metadata."""
    registry = DeviceRegistry()

    def filter_func(entity) -> bool:
        return entity.__class__.__name__ == "LQISensor"

    (
        QuirkBuilder("CentraLite", "3405-L", registry=registry)
        .change_entity_metadata(
            endpoint_id=1,
            unique_id_suffix="-lqi",
            new_device_class=SensorDeviceClass.POWER,
            new_state_class=SensorStateClass.MEASUREMENT,
            new_entity_category=EntityType.CONFIG,
            new_entity_registry_enabled_default=False,
            new_translation_key="custom_lqi_key",
            new_translation_placeholders={"placeholder": "123"},
            new_fallback_name="Custom LQI Name",
        )
        .change_entity_metadata(
            function=filter_func,
            new_unique_id="custom_lqi_unique_id",
        )
        .change_entity_metadata(
            endpoint_id=1,
            cluster_id=general.Identify.cluster_id,
            cluster_type=ClusterType.Server,
            new_primary=True,
        )
        .add_to_registry()
    )

    zigpy_dev = registry.get_device(
        await zigpy_device_from_json(
            zha_gateway.application_controller,
            "tests/data/devices/centralite-3405-l.json",
        )
    )

    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    # Find the LQI sensor entity to verify metadata changes were applied
    lqi_entity = get_entity(
        zha_device, platform=Platform.SENSOR, qualifier="custom_lqi_unique_id"
    )

    assert lqi_entity is not None, "LQI sensor entity should exist"

    # Verify metadata changes were applied - first filter matches by endpoint_id=1 and unique_id_suffix="-lqi"
    assert lqi_entity._attr_device_class == SensorDeviceClass.POWER
    assert lqi_entity._attr_state_class == SensorStateClass.MEASUREMENT
    assert lqi_entity._attr_entity_category == EntityType.CONFIG
    assert lqi_entity._attr_entity_registry_enabled_default is False
    assert lqi_entity._attr_translation_key == "custom_lqi_key"
    assert lqi_entity._attr_translation_placeholders == {"placeholder": "123"}
    assert lqi_entity._attr_fallback_name == "Custom LQI Name"

    # Verify metadata changes from second filter - function-based match
    assert lqi_entity.unique_id == "custom_lqi_unique_id"

    button_entity = get_entity(zha_device, platform=Platform.BUTTON)
    assert button_entity._attr_primary is True


async def test_quirks_v2_translation_placeholders(zha_gateway: Gateway) -> None:
    """Test quirks v2 translation_placeholders on entities."""
    registry = DeviceRegistry()

    (
        QuirkBuilder("CentraLite", "3405-L", registry=registry)
        .sensor(
            PowerConfiguration.AttributeDefs.battery_voltage.name,
            PowerConfiguration.cluster_id,
            translation_key="some_battery_sensor",
            translation_placeholders={"sensor_index": "1"},
            fallback_name="Some battery sensor {sensor_index}",
        )
        .add_to_registry()
    )

    zigpy_dev = registry.get_device(
        await zigpy_device_from_json(
            zha_gateway.application_controller,
            "tests/data/devices/centralite-3405-l.json",
        )
    )

    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)
    entity = get_entity(
        zha_device, platform=Platform.SENSOR, qualifier="battery_voltage"
    )
    assert entity is not None

    assert (
        entity.translation_placeholders
        == entity.info_object.translation_placeholders
        == {"sensor_index": "1"}
    )


async def test_quirks_v2_exposed_features(zha_gateway: Gateway) -> None:
    """Test quirks v2 exposed features."""
    registry = DeviceRegistry()

    (
        QuirkBuilder("CentraLite", "3405-L", registry=registry)
        .exposes_feature("some_feature")
        .exposes_feature("another_feature", config={"option": True})
        .add_to_registry()
    )

    zigpy_dev = registry.get_device(
        await zigpy_device_from_json(
            zha_gateway.application_controller,
            "tests/data/devices/centralite-3405-l.json",
        )
    )

    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    # can access the set of exposed features, similar to v1 quirks
    assert zha_device.exposes_features == {"some_feature", "another_feature"}

    # can access the quirk metadata for config features
    assert zha_device.quirk_metadata.exposes_features == (
        ExposesFeatureMetadata(feature="some_feature"),
        ExposesFeatureMetadata(feature="another_feature", config={"option": True}),
    )


async def test_join_binding_reporting(zha_gateway: Gateway) -> None:
    """Test that new joins go through binding and attribute reporting."""

    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/espressif-zigbeecarbondioxidesensor.json",
    )

    co2 = zigpy_dev.endpoints[10].carbon_dioxide_concentration

    with (
        patch.object(co2, "bind", wraps=co2.bind) as mock_bind,
        patch.object(
            co2, "configure_reporting_multiple", wraps=co2.configure_reporting_multiple
        ) as mock_reporting_config,
    ):
        await join_zigpy_device(zha_gateway, zigpy_dev)

    assert mock_bind.mock_calls == [call()]
    assert mock_reporting_config.mock_calls == [
        call({"measured_value": (30, 900, 1e-6)})
    ]


async def test_endpoint_none_profile(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test endpoint with None profile id being skipped."""
    zigpy_dev = zigpy_device(zha_gateway, with_basic_cluster_handler=True)
    zigpy_dev.endpoints[3].profile_id = None
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    assert zha_device.async_get_std_clusters() == {}
    assert "Skipping endpoint, profile is None" in caplog.text


async def test_somrig_events(zha_gateway: Gateway) -> None:
    """Test that Somrig events are handled correctly."""

    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/ikea-of-sweden-somrig-shortcut-button.json",
    )

    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    listener = mock.Mock()
    zha_device.on_all_events(listener)

    # ShortcutV2Cluster:initial_press(new_position=0)
    zigpy_dev.packet_received(
        zigpy.types.ZigbeePacket(
            src_ep=1,
            dst_ep=1,
            tsn=0,
            profile_id=260,
            cluster_id=64640,
            data=zigpy.types.SerializableBytes(b"\x15|\x11\x0f\x01\x00"),
        )
    )

    assert listener.mock_calls == [
        call(
            ZHAEvent(
                device_ieee=zigpy.types.EUI64.convert("ab:cd:ef:12:6d:e6:02:47"),
                unique_id="ab:cd:ef:12:6d:e6:02:47",
                data={
                    "unique_id": "ab:cd:ef:12:6d:e6:02:47:1:0xfc80_CLIENT",
                    "endpoint_id": 1,
                    "cluster_id": 64640,
                    "command": "initial_press",
                    "args": [0],
                    "params": {"new_position": 0},
                },
                event_type="zha_event",
                event="zha_event",
            )
        )
    ]


async def test_symfonisk_events(
    zha_gateway: Gateway,
) -> None:
    """Test that Symfonisk events are handled correctly."""

    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/ikea-of-sweden-symfonisk-sound-remote-gen2.json",
    )

    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    listener = mock.Mock()
    zha_device.on_all_events(listener)

    # ShortcutV1Cluster:shortcut_v1_events(shortcut_button=1, shortcut_event=1)
    zigpy_dev.packet_received(
        zigpy.types.ZigbeePacket(
            src_ep=1,
            dst_ep=1,
            tsn=0,
            profile_id=260,
            cluster_id=64639,
            data=zigpy.types.SerializableBytes(b"\x15|\x11\x1f\x01\x01\x01"),
        )
    )

    assert listener.mock_calls == [
        call(
            ZHAEvent(
                device_ieee=zigpy.types.EUI64.convert("ab:cd:ef:12:52:61:2b:43"),
                unique_id="ab:cd:ef:12:52:61:2b:43",
                data={
                    "unique_id": "ab:cd:ef:12:52:61:2b:43:1:0xfc7f_CLIENT",
                    "endpoint_id": 1,
                    "cluster_id": 64639,
                    "command": "shortcut_v1_events",
                    "args": [1, 1],
                    "params": {"shortcut_button": 1, "shortcut_event": 1},
                },
                event_type="zha_event",
                event="zha_event",
            )
        )
    ]


async def test_device_on_remove_callback_failure(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test that device.on_remove continues when callback fails."""
    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/philips-sml001.json",
    )
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    failing_callback = mock.Mock(side_effect=Exception("Callback failed"))
    zha_device._on_remove_callbacks.append(failing_callback)

    await zha_device.on_remove()

    assert failing_callback.call_count == 1
    assert "Failed to execute on_remove callback" in caplog.text
    assert "Callback failed" in caplog.text


async def test_device_on_remove_platform_entity_failure(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test that device.on_remove continues when platform entity removal fails."""
    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/philips-sml001.json",
    )
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    switch_entity = get_entity(zha_device, platform=Platform.SWITCH)
    with patch.object(
        switch_entity, "on_remove", side_effect=Exception("Entity removal failed")
    ):
        await zha_device.on_remove()

    assert "Failed to remove platform entity" in caplog.text
    assert "Entity removal failed" in caplog.text


async def test_device_on_remove_pending_entity_failure(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test that device.on_remove continues when pending entity removal fails."""
    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/philips-sml001.json",
    )
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    update_entity = get_entity(zha_device, platform=Platform.UPDATE)
    zha_device._pending_entities.append(update_entity)

    with patch.object(
        update_entity,
        "on_remove",
        side_effect=Exception("Pending entity removal failed"),
    ):
        await zha_device.on_remove()

    assert "Failed to remove pending entity" in caplog.text
    assert "Pending entity removal failed" in caplog.text



"""Support for ZHA button."""

from __future__ import annotations

from dataclasses import dataclass
import functools
import logging
from typing import TYPE_CHECKING, Any

from zigpy.quirks.v2 import WriteAttributeButtonMetadata, ZCLCommandButtonMetadata
from zigpy.zcl.clusters.general import Identify

from zha.application import Platform
from zha.application.const import ENTITY_METADATA
from zha.application.platforms import (
    BaseEntity,
    BaseEntityInfo,
    ClusterHandlerMatch,
    EntityCategory,
    PlatformEntity,
    register_entity,
)
from zha.application.platforms.button.const import DEFAULT_DURATION, ButtonDeviceClass
from zha.zigbee.cluster_handlers.const import (
    AQARA_OPPLE_CLUSTER,
    CLUSTER_HANDLER_IDENTIFY,
    TUYA_MANUFACTURER_CLUSTER,
)

if TYPE_CHECKING:
    from zha.zigbee.cluster_handlers import ClusterHandler
    from zha.zigbee.device import Device
    from zha.zigbee.endpoint import Endpoint

_LOGGER = logging.getLogger(__name__)


@dataclass(frozen=True, kw_only=True)
class CommandButtonEntityInfo(BaseEntityInfo):
    """Command button entity info."""

    command: str
    args: list[Any]
    kwargs: dict[str, Any]


@dataclass(frozen=True, kw_only=True)
class WriteAttributeButtonEntityInfo(BaseEntityInfo):
    """Write attribute button entity info."""

    attribute_name: str
    attribute_value: Any


class Button(PlatformEntity):
    """Defines a ZHA button."""

    PLATFORM = Platform.BUTTON

    _command_name: str
    _args: list[Any]
    _kwargs: dict[str, Any]

    def __init__(
        self,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
        **kwargs: Any,
    ):
        """Initialize button."""
        self._cluster_handler: ClusterHandler = cluster_handlers[0]
        if ENTITY_METADATA in kwargs:
            self._init_from_quirks_metadata(kwargs[ENTITY_METADATA])
        super().__init__(cluster_handlers, endpoint, device, **kwargs)

    def _init_from_quirks_metadata(
        self, entity_metadata: ZCLCommandButtonMetadata
    ) -> None:
        """Init this entity from the quirks metadata."""
        super()._init_from_quirks_metadata(entity_metadata)
        self._command_name = entity_metadata.command_name
        self._args = entity_metadata.args
        self._kwargs = entity_metadata.kwargs

    @functools.cached_property
    def info_object(self) -> CommandButtonEntityInfo:
        """Return a representation of the button."""
        return CommandButtonEntityInfo(
            **super().info_object.__dict__,
            command=self._command_name,
            args=self._args,
            kwargs=self._kwargs,
        )

    @functools.cached_property
    def args(self) -> list[Any]:
        """Return the arguments to use in the command."""
        return list(self._args) if self._args else []

    @functools.cached_property
    def kwargs(self) -> dict[str, Any]:
        """Return the keyword arguments to use in the command."""
        return self._kwargs

    async def async_press(self) -> None:
        """Send out a update command."""
        command = getattr(self._cluster_handler, self._command_name)
        arguments = self.args or []
        kwargs = self.kwargs or {}
        await command(*arguments, **kwargs)


@register_entity(Identify.cluster_id)
class IdentifyButton(Button):
    """Defines a ZHA identify button."""

    _attr_device_class = ButtonDeviceClass.IDENTIFY
    _attr_entity_category = EntityCategory.DIAGNOSTIC
    _command_name = "identify"
    _kwargs = {}
    _args = [DEFAULT_DURATION]

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_IDENTIFY})
    )

    def is_supported_in_list(self, entities: list[BaseEntity]) -> bool:
        """Check if this button is supported given the list of entities."""
        cls = type(self)
        return not any(type(entity) is cls for entity in entities)


class WriteAttributeButton(PlatformEntity):
    """Defines a ZHA button, which writes a value to an attribute."""

    PLATFORM = Platform.BUTTON

    _attribute_name: str
    _attribute_value: Any = None

    def __init__(
        self,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
        **kwargs: Any,
    ) -> None:
        """Init this button."""
        self._cluster_handler: ClusterHandler = cluster_handlers[0]
        if ENTITY_METADATA in kwargs:
            self._init_from_quirks_metadata(kwargs[ENTITY_METADATA])
        super().__init__(cluster_handlers, endpoint, device, **kwargs)
        self.recompute_capabilities()

    def _init_from_quirks_metadata(
        self, entity_metadata: WriteAttributeButtonMetadata
    ) -> None:
        """Init this entity from the quirks metadata."""
        super()._init_from_quirks_metadata(entity_metadata)
        self._attribute_name = entity_metadata.attribute_name
        self._attribute_value = entity_metadata.attribute_value

    @functools.cached_property
    def info_object(self) -> WriteAttributeButtonEntityInfo:
        """Return a representation of the button."""
        return WriteAttributeButtonEntityInfo(
            **super().info_object.__dict__,
            attribute_name=self._attribute_name,
            attribute_value=self._attribute_value,
        )

    async def async_press(self) -> None:
        """Write attribute with defined value."""
        await self._cluster_handler.write_attributes_safe(
            {self._attribute_name: self._attribute_value}
        )


@register_entity(TUYA_MANUFACTURER_CLUSTER)
class FrostLockResetButton(WriteAttributeButton):
    """Defines a ZHA frost lock reset button."""

    _unique_id_suffix = "reset_frost_lock"
    _attribute_name = "frost_lock_reset"
    _attribute_value = 0
    _attr_device_class = ButtonDeviceClass.RESTART
    _attr_entity_category = EntityCategory.CONFIG
    _attr_translation_key = "reset_frost_lock"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"tuya_manufacturer"}),
        manufacturers=frozenset({"_TZE200_htnnfasr"}),
    )


@register_entity(AQARA_OPPLE_CLUSTER)
class NoPresenceStatusResetButton(WriteAttributeButton):
    """Defines a ZHA no presence status reset button."""

    _unique_id_suffix = "reset_no_presence_status"
    _attribute_name = "reset_no_presence_status"
    _attribute_value = 1
    _attr_device_class = ButtonDeviceClass.RESTART
    _attr_entity_category = EntityCategory.CONFIG
    _attr_translation_key = "reset_no_presence_status"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"lumi.motion.ac01"}),
    )


@register_entity(AQARA_OPPLE_CLUSTER)
class AqaraPetFeederFeedButton(WriteAttributeButton):
    """Defines a feed button for the aqara c1 pet feeder."""

    _unique_id_suffix = "feeding"
    _attribute_name = "feeding"
    _attribute_value = 1
    _attr_translation_key = "feed"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"aqara.feeder.acn001"}),
    )


@register_entity(AQARA_OPPLE_CLUSTER)
class AqaraSelfTestButton(WriteAttributeButton):
    """Defines a ZHA self-test button for Aqara smoke sensors."""

    _unique_id_suffix = "self_test"
    _attribute_name = "self_test"
    _attribute_value = 1
    _attr_entity_category = EntityCategory.CONFIG
    _attr_translation_key = "self_test"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"lumi.sensor_smoke.acn03"}),
    )



"""Switches on Zigbee Home Automation networks."""

from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
import functools
import logging
from typing import TYPE_CHECKING, Any, cast

from zhaquirks.quirk_ids import DANFOSS_ALLY_THERMOSTAT, TUYA_PLUG_ONOFF
from zigpy.profiles import zha, zll
from zigpy.quirks.v2 import SwitchMetadata
from zigpy.zcl.clusters.closures import ConfigStatus, WindowCovering, WindowCoveringMode
from zigpy.zcl.clusters.general import Basic, BinaryOutput, OnOff
from zigpy.zcl.clusters.hvac import Thermostat
from zigpy.zcl.foundation import Status

from zha.application import Platform
from zha.application.platforms import (
    BaseEntity,
    BaseEntityInfo,
    ClusterHandlerMatch,
    EntityCategory,
    GroupEntity,
    PlatformEntity,
    PlatformFeatureGroup,
    register_entity,
    register_group_entity,
)
from zha.zigbee.cluster_handlers import ClusterAttributeUpdatedEvent
from zha.zigbee.cluster_handlers.const import (
    AQARA_OPPLE_CLUSTER,
    CLUSTER_HANDLER_ATTRIBUTE_UPDATED,
    CLUSTER_HANDLER_BASIC,
    CLUSTER_HANDLER_BINARY_OUTPUT,
    CLUSTER_HANDLER_COVER,
    CLUSTER_HANDLER_INOVELLI,
    CLUSTER_HANDLER_ON_OFF,
    CLUSTER_HANDLER_THERMOSTAT,
    IKEA_AIR_PURIFIER_CLUSTER,
    INOVELLI_CLUSTER,
    SINOPE_MANUFACTURER_CLUSTER,
    TUYA_MANUFACTURER_CLUSTER,
)
from zha.zigbee.cluster_handlers.general import (
    BinaryOutputClusterHandler,
    OnOffClusterHandler,
)
from zha.zigbee.group import Group

if TYPE_CHECKING:
    from zha.zigbee.cluster_handlers import ClusterHandler
    from zha.zigbee.device import Device
    from zha.zigbee.endpoint import Endpoint

_LOGGER = logging.getLogger(__name__)


@dataclass(frozen=True, kw_only=True)
class ConfigurableAttributeSwitchInfo(BaseEntityInfo):
    """Switch configuration entity info."""

    attribute_name: str
    invert_attribute_name: str | None
    force_inverted: bool
    off_value: int
    on_value: int


class BaseSwitch(BaseEntity, ABC):
    """Common base class for zhawss switches."""

    PLATFORM = Platform.SWITCH
    _attr_primary_weight = 10

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ):
        """Initialize the switch."""
        self._on_off_cluster_handler: OnOffClusterHandler
        super().__init__(*args, **kwargs)

    @property
    def state(self) -> dict[str, Any]:
        """Return the state of the switch."""
        response = super().state
        response["state"] = self.is_on
        return response

    @property
    def is_on(self) -> bool:
        """Return if the switch is on based on the statemachine."""
        if self._on_off_cluster_handler.on_off is None:
            return False
        return self._on_off_cluster_handler.on_off

    # TODO revert this once group entities use cluster handlers
    async def async_turn_on(self, **kwargs: Any) -> None:  # pylint: disable=unused-argument
        """Turn the entity on."""
        await self._on_off_cluster_handler.turn_on()
        self.maybe_emit_state_changed_event()

    async def async_turn_off(self, **kwargs: Any) -> None:  # pylint: disable=unused-argument
        """Turn the entity off."""
        await self._on_off_cluster_handler.turn_off()
        self.maybe_emit_state_changed_event()


@register_entity(OnOff.cluster_id)
class Switch(PlatformEntity, BaseSwitch):
    """ZHA switch."""

    _attr_translation_key = "switch"
    _attr_primary_weight = 10
    _attribute_name = OnOff.AttributeDefs.on_off.name

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_ON_OFF}),
        # Switch entities have the lowest priority
        feature_priority=(PlatformFeatureGroup.LIGHT_OR_SWITCH_OR_SHADE, -1),
    )

    def __init__(
        self,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
        **kwargs: Any,
    ) -> None:
        """Initialize the ZHA switch."""
        legacy_discovery_unique_id = (
            f"{endpoint.device.ieee}-{endpoint.id}"
            if (
                endpoint.zigpy_endpoint.profile_id,
                endpoint.zigpy_endpoint.device_type,
            )
            in {
                (zha.PROFILE_ID, zha.DeviceType.ON_OFF_BALLAST),
                (zha.PROFILE_ID, zha.DeviceType.ON_OFF_PLUG_IN_UNIT),
                (zha.PROFILE_ID, zha.DeviceType.SMART_PLUG),
                (zll.PROFILE_ID, zll.DeviceType.ON_OFF_PLUGIN_UNIT),
            }
            else f"{endpoint.device.ieee}-{endpoint.id}-{int(OnOff.cluster_id)}"
        )

        super().__init__(
            cluster_handlers,
            endpoint,
            device,
            **kwargs,
            legacy_discovery_unique_id=legacy_discovery_unique_id,
        )
        self._on_off_cluster_handler: OnOffClusterHandler = cast(
            OnOffClusterHandler, self.cluster_handlers[CLUSTER_HANDLER_ON_OFF]
        )

    def on_add(self) -> None:
        """Run when entity is added."""
        super().on_add()
        self._on_remove_callbacks.append(
            self._on_off_cluster_handler.on_event(
                CLUSTER_HANDLER_ATTRIBUTE_UPDATED,
                self.handle_cluster_handler_attribute_updated,
            )
        )

    def _is_supported(self) -> bool:
        if self._on_off_cluster_handler.cluster.is_attribute_unsupported(
            self._attribute_name
        ):
            _LOGGER.debug(
                "%s is not supported - skipping %s entity creation",
                self._attribute_name,
                self.__class__.__name__,
            )
            return False

        return super()._is_supported()

    def handle_cluster_handler_attribute_updated(
        self,
        event: ClusterAttributeUpdatedEvent,  # pylint: disable=unused-argument
    ) -> None:
        """Handle state update from cluster handler."""
        if event.attribute_name == self._attribute_name:
            self.maybe_emit_state_changed_event()


@register_entity(BinaryOutput.cluster_id)
class BinaryOutputSwitch(PlatformEntity, BaseSwitch):
    """BinaryOutputCluster switch."""

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_BINARY_OUTPUT})
    )

    def __init__(
        self,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
        **kwargs: Any,
    ) -> None:
        """Initialize the switch."""
        super().__init__(cluster_handlers, endpoint, device, **kwargs)
        self._binary_output_cluster_handler: BinaryOutputClusterHandler = cast(
            BinaryOutputClusterHandler,
            self.cluster_handlers[CLUSTER_HANDLER_BINARY_OUTPUT],
        )

    def _is_supported(self) -> bool:
        if self._binary_output_cluster_handler.description is None:
            return False

        return super()._is_supported()

    def recompute_capabilities(self) -> None:
        """Recompute capabilities."""
        super().recompute_capabilities()
        self._attr_fallback_name = self._binary_output_cluster_handler.description

    def on_add(self) -> None:
        """Run when entity is added."""
        super().on_add()
        self._on_remove_callbacks.append(
            self._binary_output_cluster_handler.on_event(
                CLUSTER_HANDLER_ATTRIBUTE_UPDATED,
                self.handle_cluster_handler_attribute_updated,
            )
        )

    @property
    def is_on(self) -> bool:
        """Return if the switch is on."""
        if self._binary_output_cluster_handler.present_value is None:
            return False
        return bool(self._binary_output_cluster_handler.present_value)

    async def async_turn_on(self, **kwargs: Any) -> None:  # pylint: disable=unused-argument
        """Turn the entity on."""
        await self._binary_output_cluster_handler.async_set_present_value(True)
        self.maybe_emit_state_changed_event()

    async def async_turn_off(self, **kwargs: Any) -> None:  # pylint: disable=unused-argument
        """Turn the entity off."""
        await self._binary_output_cluster_handler.async_set_present_value(False)
        self.maybe_emit_state_changed_event()

    def handle_cluster_handler_attribute_updated(
        self,
        event: ClusterAttributeUpdatedEvent,  # pylint: disable=unused-argument
    ) -> None:
        """Handle state update from cluster handler."""
        if event.attribute_name == BinaryOutput.AttributeDefs.present_value.name:
            self.maybe_emit_state_changed_event()


@register_group_entity
class SwitchGroup(GroupEntity, BaseSwitch):
    """Representation of a switch group."""

    def __init__(self, group: Group):
        """Initialize a switch group."""
        super().__init__(group)
        self._state: bool
        self._on_off_cluster_handler = group.zigpy_group.endpoint[OnOff.cluster_id]
        if hasattr(self, "info_object"):
            delattr(self, "info_object")
        self.update()

    @property
    def is_on(self) -> bool:
        """Return if the switch is on based on the statemachine."""
        return bool(self._state)

    async def async_turn_on(self, **kwargs: Any) -> None:  # pylint: disable=unused-argument
        """Turn the entity on."""
        result = await self._on_off_cluster_handler.on()
        if isinstance(result, Exception) or result[1] is not Status.SUCCESS:
            return
        self._state = True
        self.maybe_emit_state_changed_event()

    async def async_turn_off(self, **kwargs: Any) -> None:  # pylint: disable=unused-argument
        """Turn the entity off."""
        result = await self._on_off_cluster_handler.off()
        if isinstance(result, Exception) or result[1] is not Status.SUCCESS:
            return
        self._state = False
        self.maybe_emit_state_changed_event()

    def update(self, _: Any | None = None) -> None:
        """Query all members and determine the light group state."""
        self.debug("Updating switch group entity state")
        platform_entities = self._group.get_platform_entities(self.PLATFORM)
        all_states = [entity.state for entity in platform_entities]
        self.debug(
            "All platform entity states for group entity members: %s", all_states
        )
        on_states = [state for state in all_states if state["state"]]

        self._state = len(on_states) > 0

        self.maybe_emit_state_changed_event()


class ConfigurableAttributeSwitch(PlatformEntity):
    """Representation of a ZHA switch configuration entity."""

    PLATFORM = Platform.SWITCH

    _attr_entity_category = EntityCategory.CONFIG
    _attribute_name: str
    _inverter_attribute_name: str | None = None
    _force_inverted: bool = False
    _off_value: int = 0
    _on_value: int = 1

    def __init__(
        self,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
        *,
        legacy_discovery_unique_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Init this number configuration entity."""
        self._cluster_handler: ClusterHandler = cluster_handlers[0]

        if legacy_discovery_unique_id is None:
            legacy_discovery_unique_id = (
                f"{endpoint.device.ieee}-{endpoint.id}"
                if (
                    endpoint.zigpy_endpoint.profile_id,
                    endpoint.zigpy_endpoint.device_type,
                )
                in {
                    (zha.PROFILE_ID, zha.DeviceType.ON_OFF_BALLAST),
                    (zha.PROFILE_ID, zha.DeviceType.ON_OFF_PLUG_IN_UNIT),
                    (zha.PROFILE_ID, zha.DeviceType.SMART_PLUG),
                    (zll.PROFILE_ID, zll.DeviceType.ON_OFF_PLUGIN_UNIT),
                }
                else f"{endpoint.device.ieee}-{endpoint.id}-{int(cluster_handlers[0].cluster.cluster_id)}"
            )

        super().__init__(
            cluster_handlers,
            endpoint,
            device,
            **kwargs,
            legacy_discovery_unique_id=legacy_discovery_unique_id,
        )
        self._cluster_handler.on_event(
            CLUSTER_HANDLER_ATTRIBUTE_UPDATED,
            self.handle_cluster_handler_attribute_updated,
        )

    def _init_from_quirks_metadata(self, entity_metadata: SwitchMetadata) -> None:
        """Init this entity from the quirks metadata."""
        super()._init_from_quirks_metadata(entity_metadata)
        self._attribute_name = entity_metadata.attribute_name
        if entity_metadata.invert_attribute_name:
            self._inverter_attribute_name = entity_metadata.invert_attribute_name
        if entity_metadata.force_inverted:
            self._force_inverted = entity_metadata.force_inverted
        self._off_value = entity_metadata.off_value
        self._on_value = entity_metadata.on_value

    def _is_supported(self) -> bool:
        if (
            self._attribute_name not in self._cluster_handler.cluster.attributes_by_name
            or self._cluster_handler.cluster.is_attribute_unsupported(
                self._attribute_name
            )
            or self._cluster_handler.cluster.get(self._attribute_name) is None
        ):
            _LOGGER.debug(
                "%s is not supported - skipping %s entity creation",
                self._attribute_name,
                self.__class__.__name__,
            )
            return False

        return super()._is_supported()

    @functools.cached_property
    def info_object(self) -> ConfigurableAttributeSwitchInfo:
        """Return representation of the switch configuration entity."""
        return ConfigurableAttributeSwitchInfo(
            **super().info_object.__dict__,
            attribute_name=self._attribute_name,
            invert_attribute_name=self._inverter_attribute_name,
            force_inverted=self._force_inverted,
            off_value=self._off_value,
            on_value=self._on_value,
        )

    @property
    def state(self) -> dict[str, Any]:
        """Return the state of the switch."""
        response = super().state
        response["state"] = self.is_on
        response["inverted"] = self.inverted
        return response

    @property
    def inverted(self) -> bool:
        """Return True if the switch is inverted."""
        if self._inverter_attribute_name:
            return bool(
                self._cluster_handler.cluster.get(self._inverter_attribute_name)
            )
        return self._force_inverted

    @property
    def is_on(self) -> bool:
        """Return if the switch is on based on the statemachine."""
        if self._on_value != 1:
            val = self._cluster_handler.cluster.get(self._attribute_name)
            val = val == self._on_value
        else:
            val = bool(self._cluster_handler.cluster.get(self._attribute_name))
        return (not val) if self.inverted else val

    def handle_cluster_handler_attribute_updated(
        self,
        event: ClusterAttributeUpdatedEvent,  # pylint: disable=unused-argument
    ) -> None:
        """Handle state update from cluster handler."""
        if event.attribute_name == self._attribute_name:
            self.maybe_emit_state_changed_event()

    async def async_turn_on_off(self, state: bool) -> None:
        """Turn the entity on or off."""
        if self.inverted:
            state = not state
        if state:
            await self._cluster_handler.write_attributes_safe(
                {self._attribute_name: self._on_value}
            )
        else:
            await self._cluster_handler.write_attributes_safe(
                {self._attribute_name: self._off_value}
            )
        self.maybe_emit_state_changed_event()

    async def async_turn_on(self, **kwargs: Any) -> None:  # pylint: disable=unused-argument
        """Turn the entity on."""
        await self.async_turn_on_off(True)

    async def async_turn_off(self, **kwargs: Any) -> None:  # pylint: disable=unused-argument
        """Turn the entity off."""
        await self.async_turn_on_off(False)

    async def async_update(self) -> None:
        """Attempt to retrieve the state of the entity."""
        self.debug("Polling current state")

        polling_attrs = [self._attribute_name]
        if self._inverter_attribute_name:
            polling_attrs.append(self._inverter_attribute_name)

        results = await self._cluster_handler.get_attributes(
            polling_attrs, from_cache=False, only_cache=False
        )

        self.debug("read values=%s", results)
        self.maybe_emit_state_changed_event()


@register_entity(TUYA_MANUFACTURER_CLUSTER)
class OnOffWindowDetectionFunctionConfigurationEntity(ConfigurableAttributeSwitch):
    """Representation of a ZHA window detection configuration entity."""

    _unique_id_suffix = "on_off_window_opened_detection"
    _attribute_name = "window_detection_function"
    _inverter_attribute_name = "window_detection_function_inverter"
    _attr_translation_key = "window_detection_function"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"tuya_manufacturer"}),
        manufacturers=frozenset({"_TZE200_b6wax7g0"}),
    )


@register_entity(AQARA_OPPLE_CLUSTER)
class P1MotionTriggerIndicatorSwitch(ConfigurableAttributeSwitch):
    """Representation of a ZHA motion triggering configuration entity."""

    _unique_id_suffix = "trigger_indicator"
    _attribute_name = "trigger_indicator"
    _attr_translation_key = "trigger_indicator"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"lumi.motion.ac02"}),
    )


@register_entity(AQARA_OPPLE_CLUSTER)
class XiaomiPlugPowerOutageMemorySwitch(ConfigurableAttributeSwitch):
    """Representation of a ZHA power outage memory configuration entity."""

    _unique_id_suffix = "power_outage_memory"
    _attribute_name = "power_outage_memory"
    _attr_translation_key = "power_outage_memory"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"lumi.plug.mmeu01", "lumi.plug.maeu01"}),
    )


@register_entity(Basic.cluster_id)
class HueMotionTriggerIndicatorSwitch(ConfigurableAttributeSwitch):
    """Representation of a ZHA motion triggering configuration entity."""

    _unique_id_suffix = "trigger_indicator"
    _attribute_name = "trigger_indicator"
    _attr_translation_key = "trigger_indicator"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_BASIC}),
        manufacturers=frozenset({"Philips", "Signify Netherlands B.V."}),
        models=frozenset({"SML001", "SML002", "SML003", "SML004"}),
    )


@register_entity(IKEA_AIR_PURIFIER_CLUSTER)
class ChildLock(ConfigurableAttributeSwitch):
    """ZHA BinarySensor."""

    _unique_id_suffix = "child_lock"
    _attribute_name = "child_lock"
    _attr_translation_key = "child_lock"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"ikea_airpurifier"}),
        models=frozenset({"STARKVIND Air purifier", "STARKVIND Air purifier table"}),
    )


@register_entity(IKEA_AIR_PURIFIER_CLUSTER)
class DisableLed(ConfigurableAttributeSwitch):
    """ZHA BinarySensor."""

    _unique_id_suffix = "disable_led"
    _attribute_name = "disable_led"
    _attr_translation_key = "disable_led"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"ikea_airpurifier"}),
        models=frozenset({"STARKVIND Air purifier", "STARKVIND Air purifier table"}),
    )


@register_entity(INOVELLI_CLUSTER)
class InovelliInvertSwitch(ConfigurableAttributeSwitch):
    """Inovelli invert switch control."""

    _unique_id_suffix = "invert_switch"
    _attribute_name = "invert_switch"
    _attr_translation_key = "invert_switch"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_INOVELLI}),
    )


@register_entity(INOVELLI_CLUSTER)
class InovelliSmartBulbMode(ConfigurableAttributeSwitch):
    """Inovelli smart bulb mode control."""

    _unique_id_suffix = "smart_bulb_mode"
    _attribute_name = "smart_bulb_mode"
    _attr_translation_key = "smart_bulb_mode"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_INOVELLI}),
    )


@register_entity(INOVELLI_CLUSTER)
class InovelliSmartFanMode(ConfigurableAttributeSwitch):
    """Inovelli smart fan mode control."""

    _unique_id_suffix = "smart_fan_mode"
    _attribute_name = "smart_fan_mode"
    _attr_translation_key = "smart_fan_mode"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_INOVELLI}),
        models=frozenset({"VZM35-SN"}),
    )


@register_entity(INOVELLI_CLUSTER)
class InovelliDoubleTapUpEnabled(ConfigurableAttributeSwitch):
    """Inovelli double tap up enabled."""

    _unique_id_suffix = "double_tap_up_enabled"
    _attribute_name = "double_tap_up_enabled"
    _attr_translation_key = "double_tap_up_enabled"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_INOVELLI}),
    )


@register_entity(INOVELLI_CLUSTER)
class InovelliDoubleTapDownEnabled(ConfigurableAttributeSwitch):
    """Inovelli double tap down enabled."""

    _unique_id_suffix = "double_tap_down_enabled"
    _attribute_name = "double_tap_down_enabled"
    _attr_translation_key = "double_tap_down_enabled"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_INOVELLI}),
    )


@register_entity(INOVELLI_CLUSTER)
class InovelliAuxSwitchScenes(ConfigurableAttributeSwitch):
    """Inovelli unique aux switch scenes."""

    _unique_id_suffix = "aux_switch_scenes"
    _attribute_name = "aux_switch_scenes"
    _attr_translation_key = "aux_switch_scenes"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_INOVELLI}),
    )


@register_entity(INOVELLI_CLUSTER)
class InovelliBindingOffToOnSyncLevel(ConfigurableAttributeSwitch):
    """Inovelli send move to level with on/off to bound devices."""

    _unique_id_suffix = "binding_off_to_on_sync_level"
    _attribute_name = "binding_off_to_on_sync_level"
    _attr_translation_key = "binding_off_to_on_sync_level"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_INOVELLI}),
    )


@register_entity(INOVELLI_CLUSTER)
class InovelliLocalProtection(ConfigurableAttributeSwitch):
    """Inovelli local protection control."""

    _unique_id_suffix = "local_protection"
    _attribute_name = "local_protection"
    _attr_translation_key = "local_protection"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_INOVELLI}),
    )


@register_entity(INOVELLI_CLUSTER)
class InovelliOnOffLEDMode(ConfigurableAttributeSwitch):
    """Inovelli only 1 LED mode control."""

    _unique_id_suffix = "on_off_led_mode"
    _attribute_name = "on_off_led_mode"
    _attr_translation_key = "one_led_mode"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_INOVELLI}),
    )


@register_entity(INOVELLI_CLUSTER)
class InovelliFirmwareProgressLED(ConfigurableAttributeSwitch):
    """Inovelli firmware progress LED control."""

    _unique_id_suffix = "firmware_progress_led"
    _attribute_name = "firmware_progress_led"
    _attr_translation_key = "firmware_progress_led"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_INOVELLI}),
    )


@register_entity(INOVELLI_CLUSTER)
class InovelliRelayClickInOnOffMode(ConfigurableAttributeSwitch):
    """Inovelli relay click in on off mode control."""

    _unique_id_suffix = "relay_click_in_on_off_mode"
    _attribute_name = "relay_click_in_on_off_mode"
    _attr_translation_key = "relay_click_in_on_off_mode"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_INOVELLI}),
    )


@register_entity(INOVELLI_CLUSTER)
class InovelliDisableDoubleTapClearNotificationsMode(ConfigurableAttributeSwitch):
    """Inovelli disable clear notifications double tap control."""

    _unique_id_suffix = "disable_clear_notifications_double_tap"
    _attribute_name = "disable_clear_notifications_double_tap"
    _attr_translation_key = "disable_clear_notifications_double_tap"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_INOVELLI}),
    )


@register_entity(AQARA_OPPLE_CLUSTER)
class AqaraPetFeederLEDIndicator(ConfigurableAttributeSwitch):
    """Representation of a LED indicator configuration entity."""

    _unique_id_suffix = "disable_led_indicator"
    _attribute_name = "disable_led_indicator"
    _attr_translation_key = "led_indicator"
    _force_inverted = True

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"aqara.feeder.acn001"}),
    )


@register_entity(AQARA_OPPLE_CLUSTER)
class AqaraPetFeederChildLock(ConfigurableAttributeSwitch):
    """Representation of a child lock configuration entity."""

    _unique_id_suffix = "child_lock"
    _attribute_name = "child_lock"
    _attr_translation_key = "child_lock"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"aqara.feeder.acn001"}),
    )


@register_entity(OnOff.cluster_id)
class TuyaChildLockSwitch(ConfigurableAttributeSwitch):
    """Representation of a child lock configuration entity."""

    _unique_id_suffix = "child_lock"
    _attribute_name = "child_lock"
    _attr_translation_key = "child_lock"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_ON_OFF}),
        exposed_features=frozenset({TUYA_PLUG_ONOFF}),
    )


@register_entity(AQARA_OPPLE_CLUSTER)
class AqaraThermostatWindowDetection(ConfigurableAttributeSwitch):
    """Representation of an Aqara thermostat window detection configuration entity."""

    _unique_id_suffix = "window_detection"
    _attribute_name = "window_detection"
    _attr_translation_key = "window_detection"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"lumi.airrtc.agl001"}),
    )


@register_entity(AQARA_OPPLE_CLUSTER)
class AqaraThermostatValveDetection(ConfigurableAttributeSwitch):
    """Representation of an Aqara thermostat valve detection configuration entity."""

    _unique_id_suffix = "valve_detection"
    _attribute_name = "valve_detection"
    _attr_translation_key = "valve_detection"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"lumi.airrtc.agl001"}),
    )


@register_entity(AQARA_OPPLE_CLUSTER)
class AqaraThermostatChildLock(ConfigurableAttributeSwitch):
    """Representation of an Aqara thermostat child lock configuration entity."""

    _unique_id_suffix = "child_lock"
    _attribute_name = "child_lock"
    _attr_translation_key = "child_lock"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"lumi.airrtc.agl001"}),
    )


@register_entity(AQARA_OPPLE_CLUSTER)
class AqaraHeartbeatIndicator(ConfigurableAttributeSwitch):
    """Representation of a heartbeat indicator configuration entity for Aqara smoke sensors."""

    _unique_id_suffix = "heartbeat_indicator"
    _attribute_name = "heartbeat_indicator"
    _attr_translation_key = "heartbeat_indicator"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"lumi.sensor_smoke.acn03"}),
    )


@register_entity(AQARA_OPPLE_CLUSTER)
class AqaraLinkageAlarm(ConfigurableAttributeSwitch):
    """Representation of a linkage alarm configuration entity for Aqara smoke sensors."""

    _unique_id_suffix = "linkage_alarm"
    _attribute_name = "linkage_alarm"
    _attr_translation_key = "linkage_alarm"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"lumi.sensor_smoke.acn03"}),
    )


@register_entity(AQARA_OPPLE_CLUSTER)
class AqaraBuzzerManualMute(ConfigurableAttributeSwitch):
    """Representation of a buzzer manual mute configuration entity for Aqara smoke sensors."""

    _unique_id_suffix = "buzzer_manual_mute"
    _attribute_name = "buzzer_manual_mute"
    _attr_translation_key = "buzzer_manual_mute"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"lumi.sensor_smoke.acn03"}),
    )


@register_entity(AQARA_OPPLE_CLUSTER)
class AqaraBuzzerManualAlarm(ConfigurableAttributeSwitch):
    """Representation of a buzzer manual mute configuration entity for Aqara smoke sensors."""

    _unique_id_suffix = "buzzer_manual_alarm"
    _attribute_name = "buzzer_manual_alarm"
    _attr_translation_key = "buzzer_manual_alarm"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"lumi.sensor_smoke.acn03"}),
    )


@register_entity(WindowCovering.cluster_id)
class WindowCoveringInversionSwitch(ConfigurableAttributeSwitch):
    """Representation of a switch that controls inversion for window covering devices.

    This is necessary because this cluster uses 2 attributes to control inversion.
    """

    _unique_id_suffix = "inverted"
    _attribute_name = WindowCovering.AttributeDefs.config_status.name
    _attr_translation_key = "inverted"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_COVER}),
    )

    def _is_supported(self) -> bool:
        window_covering_mode_attr = (
            WindowCovering.AttributeDefs.window_covering_mode.name
        )

        # this entity needs a second attribute to function
        if (
            (
                self._cluster_handler.cluster.is_attribute_unsupported(
                    window_covering_mode_attr
                )
            )
            or (
                window_covering_mode_attr
                not in self._cluster_handler.cluster.attributes_by_name
            )
            or self._cluster_handler.cluster.get(window_covering_mode_attr) is None
        ):
            _LOGGER.debug(
                "%s is not supported - skipping %s entity creation",
                self._attribute_name,
                self.__class__.__name__,
            )
            return False

        return super()._is_supported()

    @property
    def is_on(self) -> bool:
        """Return if the switch is on based on the statemachine."""
        config_status = ConfigStatus(
            self._cluster_handler.cluster.get(self._attribute_name)
        )
        return ConfigStatus.Open_up_commands_reversed in config_status

    async def async_turn_on(self, **kwargs: Any) -> None:
        """Turn the entity on."""
        await self._async_on_off(True)

    async def async_turn_off(self, **kwargs: Any) -> None:
        """Turn the entity off."""
        await self._async_on_off(False)

    async def async_update(self) -> None:
        """Attempt to retrieve the state of the entity."""
        self.debug("Polling current state")
        await self._cluster_handler.get_attributes(
            [
                self._attribute_name,
                WindowCovering.AttributeDefs.window_covering_mode.name,
            ],
            from_cache=False,
            only_cache=False,
        )
        self.maybe_emit_state_changed_event()

    async def _async_on_off(self, invert: bool) -> None:
        """Turn the entity on or off."""
        name: str = WindowCovering.AttributeDefs.window_covering_mode.name
        current_mode: WindowCoveringMode = WindowCoveringMode(
            self._cluster_handler.cluster.get(name)
        )
        send_command: bool = False
        if invert and WindowCoveringMode.Motor_direction_reversed not in current_mode:
            current_mode |= WindowCoveringMode.Motor_direction_reversed
            send_command = True
        elif not invert and WindowCoveringMode.Motor_direction_reversed in current_mode:
            current_mode &= ~WindowCoveringMode.Motor_direction_reversed
            send_command = True
        if send_command:
            await self._cluster_handler.write_attributes_safe({name: current_mode})
            await self.async_update()


@register_entity(AQARA_OPPLE_CLUSTER)
class AqaraE1CurtainMotorHooksLockedSwitch(ConfigurableAttributeSwitch):
    """Representation of a switch that controls whether the curtain motor hooks are locked."""

    _unique_id_suffix = "hooks_lock"
    _attribute_name = "hooks_lock"
    _attr_translation_key = "hooks_locked"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"opple_cluster"}),
        models=frozenset({"lumi.curtain.agl001"}),
    )


@register_entity(Thermostat.cluster_id)
class DanfossExternalOpenWindowDetected(ConfigurableAttributeSwitch):
    """Danfoss proprietary attribute for communicating an open window."""

    _unique_id_suffix = "external_open_window_detected"
    _attribute_name: str = "external_open_window_detected"
    _attr_translation_key: str = "external_window_sensor"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_THERMOSTAT}),
        exposed_features=frozenset({DANFOSS_ALLY_THERMOSTAT}),
    )


@register_entity(Thermostat.cluster_id)
class DanfossWindowOpenFeature(ConfigurableAttributeSwitch):
    """Danfoss proprietary attribute enabling open window detection."""

    _unique_id_suffix = "window_open_feature"
    _attribute_name: str = "window_open_feature"
    _attr_translation_key: str = "use_internal_window_detection"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_THERMOSTAT}),
        exposed_features=frozenset({DANFOSS_ALLY_THERMOSTAT}),
    )


@register_entity(Thermostat.cluster_id)
class DanfossMountingModeControl(ConfigurableAttributeSwitch):
    """Danfoss proprietary attribute for switching to mounting mode."""

    _unique_id_suffix = "mounting_mode_control"
    _attribute_name: str = "mounting_mode_control"
    _attr_translation_key: str = "mounting_mode"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_THERMOSTAT}),
        exposed_features=frozenset({DANFOSS_ALLY_THERMOSTAT}),
    )


@register_entity(Thermostat.cluster_id)
class DanfossRadiatorCovered(ConfigurableAttributeSwitch):
    """Danfoss proprietary attribute for communicating full usage of the external temperature sensor."""

    _unique_id_suffix = "radiator_covered"
    _attribute_name: str = "radiator_covered"
    _attr_translation_key: str = "prioritize_external_temperature_sensor"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_THERMOSTAT}),
        exposed_features=frozenset({DANFOSS_ALLY_THERMOSTAT}),
    )


@register_entity(Thermostat.cluster_id)
class DanfossHeatAvailable(ConfigurableAttributeSwitch):
    """Danfoss proprietary attribute for communicating available heat."""

    _unique_id_suffix = "heat_available"
    _attribute_name: str = "heat_available"
    _attr_translation_key: str = "heat_available"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_THERMOSTAT}),
        exposed_features=frozenset({DANFOSS_ALLY_THERMOSTAT}),
    )


@register_entity(Thermostat.cluster_id)
class DanfossLoadBalancingEnable(ConfigurableAttributeSwitch):
    """Danfoss proprietary attribute for enabling load balancing."""

    _unique_id_suffix = "load_balancing_enable"
    _attribute_name: str = "load_balancing_enable"
    _attr_translation_key: str = "use_load_balancing"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_THERMOSTAT}),
        exposed_features=frozenset({DANFOSS_ALLY_THERMOSTAT}),
    )


@register_entity(Thermostat.cluster_id)
class DanfossAdaptationRunSettings(ConfigurableAttributeSwitch):
    """Danfoss proprietary attribute for enabling daily adaptation run.

    Actually a bitmap, but only the first bit is used.
    """

    _unique_id_suffix = "adaptation_run_settings"
    _attribute_name: str = "adaptation_run_settings"
    _attr_translation_key: str = "adaptation_run_enabled"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_THERMOSTAT}),
        exposed_features=frozenset({DANFOSS_ALLY_THERMOSTAT}),
    )


@register_entity(SINOPE_MANUFACTURER_CLUSTER)
class SinopeLightDoubleTapFullSwitch(ConfigurableAttributeSwitch):
    """Representation of a config option that controls whether Double Tap Full option is enabled on a Sinope light switch."""

    _unique_id_suffix = "double_up_full"
    _attribute_name = "double_up_full"
    _attr_translation_key: str = "double_up_full"

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({"sinope_manufacturer_specific"}),
        models=frozenset({"DM2500ZB", "DM2500ZB-G2", "DM2550ZB", "DM2550ZB-G2"}),
    )



"""Entity helpers for the zhaws server."""

from __future__ import annotations

from collections.abc import Callable, Iterator
import enum
import logging
from typing import TYPE_CHECKING, Any, overload

if TYPE_CHECKING:
    from zha.application.platforms.binary_sensor.const import BinarySensorDeviceClass
    from zha.application.platforms.number.const import NumberDeviceClass
    from zha.application.platforms.sensor.const import SensorDeviceClass


def find_state_attributes(states: list[dict], key: str) -> Iterator[Any]:
    """Find attributes with matching key from states."""
    for state in states:
        if (value := state.get(key)) is not None:
            yield value


def mean_int(*args: Any) -> int:
    """Return the mean of the supplied values."""
    return int(sum(args) / len(args))


def mean_tuple(*args: Any) -> tuple:
    """Return the mean values along the columns of the supplied values."""
    return tuple(sum(x) / len(x) for x in zip(*args))


def reduce_attribute(
    states: list[dict],
    key: str,
    default: Any | None = None,
    reduce: Callable[..., Any] = mean_int,
) -> Any:
    """Find the first attribute matching key from states.

    If none are found, return default.
    """
    attrs = list(find_state_attributes(states, key))

    if not attrs:
        return default

    if len(attrs) == 1:
        return attrs[0]

    return reduce(*attrs)


@overload
def validate_device_class(
    device_class_enum: type[BinarySensorDeviceClass],
    metadata_value: enum.Enum,
    platform: str,
    logger: logging.Logger,
) -> BinarySensorDeviceClass | None: ...


@overload
def validate_device_class(
    device_class_enum: type[SensorDeviceClass],
    metadata_value: enum.Enum,
    platform: str,
    logger: logging.Logger,
) -> SensorDeviceClass | None: ...


@overload
def validate_device_class(
    device_class_enum: type[NumberDeviceClass],
    metadata_value: enum.Enum,
    platform: str,
    logger: logging.Logger,
) -> NumberDeviceClass | None: ...


def validate_device_class(
    device_class_enum: (
        type[BinarySensorDeviceClass]
        | type[SensorDeviceClass]
        | type[NumberDeviceClass]
    ),
    metadata_value: enum.Enum,
    platform: str,
    logger: logging.Logger,
) -> BinarySensorDeviceClass | SensorDeviceClass | NumberDeviceClass | None:
    """Validate and return a device class."""
    try:
        return device_class_enum(metadata_value.value)
    except ValueError as ex:
        logger.warning(
            "Quirks provided an invalid device class: %s for platform %s: %s",
            metadata_value,
            platform,
            ex,
        )
        return None
